/*
    pick-set is a program to take a list of files and produce a random subset
    of these that fit within a given size.

    Copyright (C) 2000 V.R.Sanders

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To conatct me use email to vince@kyllikki.fluff.org

    Main website for new versions is http://www.kyllikki.fluff.org/software

*/


/* standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

/* headers files */
#include "pick-set.h"


int print_help(control *ctrl_opts)
{
    printf("%s:Usage %s -s <size> [options]\n",
	   ctrl_opts->prog_name,
	   ctrl_opts->prog_name);
    printf("%s:Options\n",ctrl_opts->prog_name);
    printf("v                 - version\n");
    printf("h                 - help\n");
    printf("i                 - input from stdin\n");
    printf("w                 - enable weightings\n");
    printf("e <filename>      - exclude list\n");
    printf("s <size>          - size to aim for\n");
    printf("p <10ths percent> - percent of size that is aceptable (990%%)\n");
    printf("f <number>        - max number of fidgets (100)\n");
    printf("d <level>         - debug (0-off)\n");
    printf("\n");
    return 0;
}

int print_ver()
{
    printf("Version\n");
    return 0;
}

FILE * open_file(char * filename,control *ctrl_opts)
{
    FILE * ifile;

    if (ctrl_opts->debug)
	printf("Opening file %s\n",filename);

    ifile = fopen(filename,"r");
    if (ifile==NULL)
    {
	printf("%s:Unable to open input file %s\n",
	       ctrl_opts->prog_name,
	       filename);
    }
    return ifile;
}

int remove_list_from_list(item ** src_list,item ** rem_list)
{
    return 0;
}


int read_line(char * buffer,char * outbuffer,int * buffptr,FILE * infile)
{
    int read;
    int linelen;
    char * lineend=NULL;

    while((lineend==NULL) && (*buffptr<2048))
    {
	/* attempt to fill buffer with any available data */
	read=fread(&buffer[*buffptr],1,2048-*buffptr,infile);
	/* read error */
	if (read<0)
	    return 0;

	/* eof and no data left in read buffer */
	if((read==0) && (*buffptr==0))
	    return 0;

	*buffptr+=read;

	/* find line end (if we have one) */
	lineend=index(buffer,'\n');
    }

    if (lineend==NULL)
    {
	/* oh bugger we overran our block
	 * return whole current block
	 */
	lineend=&buffer[2048];
    }

    /* copy into output block */
    linelen=(lineend-buffer);
    strncpy(outbuffer,buffer,linelen);
    outbuffer[linelen]=0;
    linelen++;

    memmove(buffer,buffer+linelen,*buffptr-linelen);
    *buffptr=*buffptr-linelen;

    return linelen;
}

/* adds item to given list - returns number of items added */
int add_item(item ** chain_head,char * filename,int weighting)
{
    item * newitem;
    struct stat filestat;

    /* refuse to add items with dodgy weightings */
    if ((weighting<1) || (weighting>100))
	return 0;

    /* check we can stat the file and do not add is its not there */
    if(stat(filename,&filestat)<0)
    {
	perror(filename);
	return 0;
    }

    newitem = malloc(sizeof(item));

    newitem->filename=strdup(filename);
    newitem->item_size=filestat.st_size;
    newitem->item_weight=weighting;

    newitem->next_item=*chain_head;
    /* link item into chain head*/
    *chain_head=newitem;

    /*    printf("added new entry %s weight %d\n",filename,weighting);*/
    
    return 1;
}

/* examines input string for a percentage weight
 * returns it if found and modifies string to omit number
 */
int read_weight(char * input_str)
{
    char * num;
    int loop;
    
    num=rindex(input_str,' ');

    /* check for deliminator */
    if (num==NULL)
	return 100;

    /* check number isnt too long */
    loop=strlen(num)-1;
    if(loop>3)
	return 100;
    
    /* check for digits */
    for(;loop>1;loop--)
	if (!isdigit(num[loop]))
	    return 100;

    loop=atoi(num);

    /* terminate filename */
    num[0]=0;

    if(loop>100)
	loop=100;

    if(loop<0)
	loop=0;
 
    return loop;    
}


/* 
 * read lines from an input stream and add to a list
 * returns an item added count
 */
int read_input(FILE * ifile,item ** list_head,int read_weights)
{
    char buffer[2048];
    char linebuff[2048];
    int buffptr=0;
    int linelen=0;
    int item_count=0;
    int weighting=100;

    linelen=read_line(buffer,linebuff,&buffptr,ifile);
    while(linelen!=0)
    {
	if(read_weights)
	    weighting=read_weight(linebuff);

	item_count+=add_item(list_head,linebuff,weighting);
	linelen=read_line(buffer,linebuff,&buffptr,ifile);
    }

    return item_count;
}

item * pick_item(item * list_base,int max_num_items,int retrys,int weightings)
{
    item * chosen;
    int skip_count;
    double largest=max_num_items;
    int weigh_against;
    
    while(retrys)
    {
	chosen=list_base;

	/* if there is only one thing return it! */
	if(max_num_items<2)
	    return chosen;

	/* use high order bits from rand result to select an item */
	skip_count=1+(int) (largest*rand()/(RAND_MAX+1.0));
	weigh_against=skip_count;
	
	while ((skip_count>1) && (chosen!=NULL))
	{
	    chosen=chosen->next_item;
	    skip_count--;
	}

	/* only do weigthings tests if enabled */
	if(weightings==0)
	    return chosen;

	/* 100 means its always chosen */
	if(chosen->item_weight==100)
	    return chosen;

	/* perform random test to check weighting */
	weigh_against=1+(int) (100.00*rand()/(RAND_MAX+1.0));
	
	if(weigh_against<=chosen->item_weight)
	    return chosen;

	retrys--;
	
    }

    /* even tho it failed it entropy test we have to return something */
    return chosen;    
}

/* remove item from specified list */
int remove_item(item* chosen,item ** src_list)
{
    item * prev_item=*src_list;

    /* remove chosen from list*/
    if(prev_item==chosen)
    {
	/* need to fix up list head - which removes chosen from list */
	*src_list=prev_item->next_item;
    }
    else
    {
	while ((prev_item->next_item !=NULL) && (prev_item->next_item!=chosen))
	    prev_item=prev_item->next_item;
	if(prev_item==NULL)
	    return 1;/* chosen isnt in list */
	/* prev_item is the previos item to chosen
	 * remove chosen by skiping its link */
	prev_item->next_item=prev_item->next_item->next_item;
    }
    return 0;

}

int move_item(item * chosen,item ** src_list,item ** dst_list)
{
    if(remove_item(chosen,src_list))
	return 1;

    /* chosen is removed from src_list - add to dst_list */
    chosen->next_item = *dst_list;
    *dst_list=chosen;

    return 0;
}

int select_set(control * ctrl_opts)
{
    item * chosen;
    unsigned long long total_size=0;
    int num_fidgets=0;

    /* loop while number of last file retrys is lower than our setting and
     * current total size is less than our aceptable min total */
    while((num_fidgets < ctrl_opts->max_fidgets) &&
	  (total_size < ctrl_opts->ok_selected_size))
    {
	/* if we have run out of input files give up */
	if(ctrl_opts->num_unselected_items==0)
	    break;

	chosen=pick_item(ctrl_opts->unselected_items,
			 ctrl_opts->num_unselected_items,
			 ctrl_opts->max_fidgets,
			 ctrl_opts->weightings);

	if (chosen==NULL)
	{
	    /* debug code
	    printf("error: chosen returned NULL!!!\n");
	    chosen=ctrl_opts->unselected_items;
	    while (chosen!=NULL)
	    {
		printf("%s\n",chosen->filename);
		chosen=chosen->next_item;
	    }
	    printf("%d\n\n",ctrl_opts->num_unselected_items);
	    */
	    continue;
	}

	if(!move_item(chosen,
		      &ctrl_opts->unselected_items,
		      &ctrl_opts->selected_items))
	{
	    ctrl_opts->num_unselected_items--;
	    ctrl_opts->num_selected_items++;
	    total_size+=chosen->item_size;
	    if(ctrl_opts->debug)
		printf("adding %s\n",chosen->filename);
	}

	/* remove a random items to get selected set below max size */
	while (total_size > ctrl_opts->max_selected_size)
	{
	    /* note removal is always done without weighting */
	    chosen=pick_item(ctrl_opts->selected_items,
			     ctrl_opts->num_selected_items,
			     ctrl_opts->max_fidgets,
			     0);

	    if (chosen==NULL)
	    {
		printf("chosen returned NULL!!!\n");
		continue;
	    }

	    if(!move_item(chosen,
			  &ctrl_opts->selected_items,
			  &ctrl_opts->unselected_items))
	    {
		ctrl_opts->num_unselected_items++;
		ctrl_opts->num_selected_items--;
		total_size-=chosen->item_size;
		num_fidgets++;
		if(ctrl_opts->debug)
		    printf("removing %s\n",chosen->filename);
	    }
	}
    }
    ctrl_opts->actual_selected_size=total_size;

    return 0;
}

int write_selected(control * ctrl_opts)
{
    item * selected=ctrl_opts->selected_items;
    while (selected!=NULL)
    {
	printf("%s\n",selected->filename);
	selected=selected->next_item;
    }
    fprintf(stderr,
	    "Target Size Was %Ld bytes (aceptable %Ld bytes)\n",
	    ctrl_opts->max_selected_size,
	    ctrl_opts->ok_selected_size);

    fprintf(stderr,"Produced Size Was %Ld bytes\n",ctrl_opts->actual_selected_size);

    fprintf(stderr,"Deficit of %Ld bytes\n",(ctrl_opts->max_selected_size-ctrl_opts->actual_selected_size));

    return 0;
}

control * new_ctrl_opts()
{
    control * ctrl_opts;
    ctrl_opts=malloc(sizeof(control));

    ctrl_opts->debug=0;
    ctrl_opts->selected_items=NULL;
    ctrl_opts->num_selected_items=0;
    ctrl_opts->unselected_items=NULL;
    ctrl_opts->num_unselected_items=0;
    ctrl_opts->exclude_items=NULL;
    ctrl_opts->num_exclude_items=0;
    ctrl_opts->max_selected_size=0;
    ctrl_opts->ok_selected_size=0;
    ctrl_opts->actual_selected_size=0;
    ctrl_opts->max_fidgets=100;
    ctrl_opts->prog_name=NULL;
    ctrl_opts->weightings=0;

    return ctrl_opts;
}

int process_cmd_line(int argc,char ** argv,control * ctrl_opts)
{
    int optsel;
    extern char *optarg;
    extern int optind, opterr;
    int percent=990;
    FILE * exclude_file=NULL;
    FILE * input_file=NULL;

    ctrl_opts->prog_name=strdup(argv[0]);

    while((optsel=getopt(argc,argv,"wvhid:s:p:f:e:"))!=EOF)
    {
	switch(optsel)
	{
	case 'i':
	    input_file=stdin;
	    break;

	case 's':
	    ctrl_opts->max_selected_size=atol(optarg);
	    break;

	case 'f':
	    ctrl_opts->max_fidgets=atoi(optarg);
	    break;

	case 'd':
	    ctrl_opts->debug=atoi(optarg);
	    break;

	case 'e':
	    exclude_file=open_file(optarg,ctrl_opts);
	    if(exclude_file==NULL)
		return 1;

	    ctrl_opts->num_exclude_items=read_input(exclude_file,
						    &ctrl_opts->exclude_items,
						    0);

	    fclose(exclude_file);

	    break;

	case 'p':
	    percent=atoi(optarg);
	    break;

	case 'w':
	    ctrl_opts->weightings=1;
	    break;
	    
	case 'v':
	    print_ver();
	    return 1;

	case 'h':
	default:
	    print_help(ctrl_opts);
	    return 1;

	}
    }

    if(input_file==NULL)
    {
	/* look for filenames of list files */
	while((argc-optind)>0)
	{
	    input_file=open_file(argv[optind],ctrl_opts);
	    if(input_file==NULL)
		return 1;

	    ctrl_opts->num_unselected_items=
				    read_input(input_file,
					       &ctrl_opts->unselected_items,
					       ctrl_opts->weightings);

	    fclose(input_file);

	    optind++;
	    
	}
	
    }
    else
    {
	ctrl_opts->num_unselected_items=
				read_input(input_file,
					   &ctrl_opts->unselected_items,
					   ctrl_opts->weightings);
    }

    /* check there is something in the input list */
    if (ctrl_opts->unselected_items==NULL)	
    {	
	print_help(ctrl_opts);
	return 1;
    }


    if(ctrl_opts->max_selected_size==0)
    {
	print_help(ctrl_opts);
	return 1;
    }
    else
    {
	/* calculate aceptable size from percentage and size */
	ctrl_opts->ok_selected_size=((ctrl_opts->max_selected_size*percent)/1000);
    }

    return 0;
}

int main(int argc,char ** argv)
{
    control * ctrl_opts;
    ctrl_opts=new_ctrl_opts();

    if(process_cmd_line(argc,argv,ctrl_opts))
	return 1;

    /* seed random generator with time now */
    srand(time(NULL));

    remove_list_from_list(&ctrl_opts->unselected_items,
			  &ctrl_opts->exclude_items);

    select_set(ctrl_opts);

    write_selected(ctrl_opts);

    free(ctrl_opts);

    return 0;
}




